#include "proximityserver.h"

#include <QTcpServer>
#include <QTcpSocket>
#include <QByteArray>
#include <QDataStream>
#include <QTextStream>
#include <QKeyEvent>
#include <QCoreApplication>
#include <QSettings>

#include "proximitynetworkdataprovider.h"
#include "proximityserialdataprovider.h"
#include "proximityplayer.h"

#include "attractiveforcelocator.h"
#include "landmarclocator.h"
#include "contactforcelocator.h"
#include "referencecontactlocator.h"

/*!
  Der Konstruktor initialisiert alle privaten Variablen und erstellt somit einen Server
  mit \sa ProximityDataReceiver, \sa ProximityDistributor und allem, was dazu gehört.
  Es kann geloggt werden - wird über \param loging festgelegt.
  */
ProximityServer::ProximityServer(quint8 modus, bool log, QString sourcePath):_modus(modus), _log(log)
{
    QSettings settings;
    _dataModel = new ProximityDataModel(); // Initialisierung des Datenmodells
    _dataModel->readFromSettings(); // hier werden die Antenneninformationen und Informationen über fixierte Tags ins Model eingelesen
    _server = new QTcpServer(this);
    _packageQueue = new QQueue<PTNP::Package*>(); // in diesem Queue werden die Packete für den Versand zu den Clients zwischengespeichert
    _clients = new QList<ProximityServerClient*>(); // die Liste der verbundenen Clients
    _clientListMutex = new QMutex();
    _packageQueueMutex = new QMutex();
    _logger = 0; // logger präinitialisieren
    QTextStream cerr(stderr, QIODevice::WriteOnly);
    if(_log) { // wenn geloggt werden soll
        _logger = new ProximityLogger(); // wird ein logger erstellt
        if(!_logger->createLogFile()) { // mit entsprechender Logdatei (Name aus Datum und Uhrzeit zusammengesetzt)
            cerr<<"ps.cpp: konnte logdatei nicht erstellen / oeffnen\n"<<flush;
        } else {
#ifdef _DEV_
            QTextStream cout(stdout, QIODevice::WriteOnly);
            cout<<"ps.cpp: logdatei erstellt!\n"<<flush;
#endif
        }
    }
    _receiver = new ProximityDataReceiver(*_packageQueue, *_packageQueueMutex, _logger, *_dataModel);
    _distributor = new ProximityDistributor(*_clients, *_clientListMutex, *_packageQueue, *_packageQueueMutex, *_dataModel);
    switch(modus) { // je nach Modus wird als dataprovider ein unterschiedlicher gewählt - alle von proximityDataProvider abgeleitet:
    case NETWORK_PLAYER:
            _dataProvider = new ProximityNetworkDataProvider();
        break;
    case LOGFILE_PLAYER:
            _dataProvider = new ProximityPlayer(sourcePath);
        break;
    case LOGFILE_PLAYER_LOOP:
            _dataProvider = new ProximityPlayer(sourcePath, true);
        break;
    case SERIAL_PLAYER:
            _dataProvider = new ProximitySerialDataProvider(sourcePath);
      break;
    default:
        cerr<<"keine Datenquelle gewaehlt!\n";
        qApp->exit(2);
        break;
    }
    initLocator();
    connect(_locator, SIGNAL(sendPositionMessage(PTNP::oid,PTNP::position)),
            _receiver, SLOT(createTagPositionPackage(PTNP::oid,PTNP::position)));
    connect(_dataProvider, SIGNAL(readyRead(QByteArray*, bool)), _receiver, SLOT(processDatagram(QByteArray*, bool)));
    connect(_receiver, SIGNAL(receivedData()), _distributor, SLOT(distribute()));
    connect(_receiver, SIGNAL(newAntennaFound()), _distributor, SLOT(distributeAllAntennasAndFixedTags()));
}


/*!
  löscht alle dynamisch generierten Objekte
  */
ProximityServer::~ProximityServer() {
    delete _receiver;
    delete _distributor;
    if(_logger!=0)
        delete _logger;
    if(_player!=0)
        delete _player;
    delete _server;
    delete _packageQueue;
    delete _clients;
    delete _clientListMutex;
    delete _packageQueueMutex;
}

/*!
  Hier wird der Server gestartet.
  \return \a true wenn der Server gestartet werden konnte und auf Anfragen reagiert. \a false wenn der Server nicht gestartet werden konnte.
  */
bool ProximityServer::start()
{
#ifdef _DEV_
    QTextStream cout(stdout, QIODevice::WriteOnly);
#endif
    if(!_server->listen(QHostAddress::Any, PORT)) {
        _server->close();
        return false;
    }

    connect(_server, SIGNAL(newConnection()), this, SLOT(initConnection()));
#ifdef _DEV_
    cout << "ps.cpp: Server gestartet\n" << flush;
#endif
    return true;
}

/*!
  Dieser Slot initialisiert die Verbindung und packt den Client in die Liste.
  */
void ProximityServer::initConnection()
{
#ifdef _DEV_
    QTextStream cout(stdout, QIODevice::WriteOnly);
    cout<<"ps.cpp: Neuer Client ... "<<flush;
#endif
    ProximityServerClient* c = new ProximityServerClient(_server->nextPendingConnection(),
                                                         *_receiver,
                                                         *_dataModel);

    _clientListMutex->lock();
    _clients->append(c);
    _clientListMutex->unlock();

    connect(c, SIGNAL(disconnected()), this, SLOT(closeConnection()));
}

/*!
  Dieser Slot wird aufgerufen, unmittelbar, bevor die Verbindung geschlossen wird.
  */
void ProximityServer::closeConnection()
{
#ifdef _DEV_
    QTextStream cout(stdout, QIODevice::WriteOnly);
#endif
    ProximityServerClient* client = qobject_cast<ProximityServerClient*>(sender());
    _clients->removeOne(client);
    client->deleteLater();
#ifdef _DEV_
    cout<<"ps.cpp: schnupps - raus ist er! noch "<<_clients->size()<<" clients\n"<<flush;
#endif
}

/*!
  Diese Methode initalisiert den Locator.
  */
void ProximityServer::initLocator() {
    QSettings settings;
    QTextStream cerr(stderr, QIODevice::WriteOnly);
    QString locator = settings.value("locator", "attractiveForce").toString();
    if(locator == "LANDMARC") {
        _locator = new LANDMARCLocator(*_dataModel);
    } else if(locator == "contactForce") {
        _locator = new ContactForceLocator(*_dataModel);
    } else if(locator == "attractiveForce"){
        _locator = new AttractiveForceLocator(*_dataModel);
    } else if(locator == "referenceContact") {
        _locator = new ReferenceContactLocator(*_dataModel);
    } else {
        cerr<<"falscher oder kein Locator in Settings angegeben!\n"
              "nehme AttractiveForce-Locator"<<endl;
        _locator = new AttractiveForceLocator(*_dataModel);
    }
}
